module Gale(GaleContext,Puff(..),galeNextPuff,showPuff, galeDecryptPuff, newGaleContext, Fragment(..), Category,
    catParseNew, catParseOld, catShowNew, catShowOld, subCategory, destroyGaleContext, connectGaleContext, reconnectGaleContext,
    getFragmentString, getFragmentStrings) where

import Network
import Concurrent
import ConcurrentUtil
import IO
import IOExts
import Char(chr,ord)
import Word
import Int
import Bits
import GenUtil
import List
import Monad(when)
import Time


-- TODO - prove concurrent-correctness, make sure all network errors are accounted for.

galePort :: PortNumber
galePort = 11512
hostStrings :: String -> [String]
hostStrings s = [s, "gale." ++ s, s ++ ".gale.org."]


data GaleContext = GaleContext { channel :: Chan Puff, categories :: [(String,String)], threads :: MVar [ThreadId] }

type Category = (String,String)
    
data Puff = Puff [Category] [(String,Fragment)] | Message String 
    deriving(Show,Read)
data Fragment = FragmentText String | FragmentData [Word8] | FragmentTime ClockTime | FragmentInt Int32 | FragmentNest [(String,Fragment)]
    deriving(Show,Read)

instance Read ClockTime where


subCategory :: Category -> Category -> Bool
(x1,y1) `subCategory` (x2,y2) = x2 `isPrefixOf` x1 && y2 `isSuffixOf` y1

newGaleContext cs = do
    let cs' = map catParseNew cs
    c <- newChan 
    thds <- mapM (forkIO . connectThread c cs') (nub $ map snd cs')
    mv <- newMVar thds
    return GaleContext { channel = c, categories = cs', threads = mv}

spc [] = []
spc s = v : spc (drop 1 r) where
    (v,r) = span (/= ':') s 

connectThread c cs d  = do
    let continue = threadDelay 2000000 >> connectThread c cs d
    --writeChan c (Message $ "Attempting Connection: " ++  d)
    h' <- try $ trySeveral (map attemptConnect (hostStrings d))
    case h' of 
	Left _ -> continue
	Right h -> catch (do
	    putWord32 h 1
	    v <- readWord32 h
	    let ss = nub [c| (c,d') <- cs, d' == d]
		f c = catShowOld (c,d)
		gs = concat $ intersperse ":" (map f ss)
	    writeChan c (Message $ "Connected: " ++  d ++ " -- " ++ unwords ss)
	    putWord32 h 2
	    putWord32 h (fromIntegral $ length gs * 2)
	    putRaw h $ galeEncodeString gs
	    sequence_ $ repeat $ do
		w <- readWord32 h
		l <- readWord32 h
		d <- readRaw h (fromIntegral l)
		when (w == 0) $ do
		    let (l,d') = xdrReadUInt d
		    writeChan c $ Puff (map catParseOld $ spc $ galeDecodeString (take (fromIntegral $ l ) d'))  (decodeFragments $ drop (fromIntegral l + 4) d'))
	      $ \_ -> do 
		hClose h
		writeChan c (Message $ "Disconnected: " ++ d)
		continue

destroyGaleContext :: GaleContext -> IO ()
destroyGaleContext gc = do
    thds <- takeMVar (threads gc)
    mapM_ killThread thds
    

reconnectGaleContext gc = closeGaleContext gc >> connectGaleContext gc

closeGaleContext = destroyGaleContext
connectGaleContext gc = do
    thds <- mapM (forkIO . connectThread (channel gc) (categories gc)) (nub $ map snd (categories gc))
    putMVar (threads gc) thds

    
galeNextPuff :: GaleContext -> IO Puff
galeNextPuff gc = readChan $ channel gc

showPuff (Puff s fl) = unwords (map catShowNew s) ++ " \n" ++ concatMap (\(n,f) -> n ++ ": " ++ showFragment f ++ "\n") fl
showPuff (Message s) = s

showFragment (FragmentText s) = s
showFragment (FragmentData _) = "<DATA>"
showFragment (FragmentTime ct) = show ct
showFragment (FragmentInt x) = show x
showFragment (FragmentNest xs) = concatMap (\(n,f) -> n ++ ": " ++ showFragment f ++ "\n") xs

galeDecryptPuff :: Puff -> IO Puff 
galeDecryptPuff (Puff s [("security/signature",FragmentData xs)]) = return $ Puff s fl where
    (l,xs') = xdrReadUInt xs
    fl = (decodeFragments $ drop (fromIntegral l + 4) xs')
galeDecryptPuff x = return x
    
decodeFragments :: [Word8] -> [(String,Fragment)]
decodeFragments [] = []
decodeFragments xs = decodeFragment ft f : decodeFragments r where
    (ft,xs') = xdrReadUInt xs
    (fl,xs'') = xdrReadUInt xs'
    (f,r) = splitAt (fromIntegral fl) xs''

--decodeFragment ft f = (show ft ++ " " ++ show f,FragmentTime)
decodeFragment :: Word32 -> [Word8] -> (String, Fragment)
decodeFragment t xs = (fn,f t) where
    (fnl,xs') = xdrReadUInt xs
    (fn,xs'') = liftT2 (galeDecodeString,id) (splitAt (fromIntegral $ fnl * 2) xs')
    f 0 = FragmentText (galeDecodeString xs'')
    f 1 = FragmentData xs''
    f 2 = FragmentTime $ decodeTime xs''
    f 3 = FragmentInt (fromIntegral $ fst $ xdrReadUInt xs'')
    f 4 = FragmentNest (decodeFragments xs'')

decodeTime :: [Word8] -> ClockTime
decodeTime xs = TimeDiff {tdDay = 0, tdPicosec = 0, tdYear = 0, tdMonth = 0, tdHour = 0, tdMin = fromIntegral (t `div` 60) , tdSec = fromIntegral (t `mod` 60)} `addToClockTime` epoch where
    (t,_) = xdrReadUInt (drop 4 xs)

attemptConnect s = do
    h <- connectTo s (PortNumber galePort)
    hSetBuffering h NoBuffering
    return h


putRaw :: Handle -> [Word8] -> IO ()
putRaw h v = hPutStr h (map (chr . fromIntegral) v)

readRaw :: Handle -> Int -> IO [Word8]
readRaw h n = do 
    v <- sequence $ replicate n (hGetChar h)
    return $ map (fromIntegral . ord) v

type XDRShowS = [Word8] -> [Word8]
type XDRReadS a = [Word8] -> (a,[Word8])


xdrWriteUInt :: Word32 -> XDRShowS
xdrWriteUInt x bs = (b1:b2:b3:b4:bs) where
    b1 = fromIntegral $ (x `shiftR` 24) .&. 0xFF
    b2 = fromIntegral $ (x `shiftR` 16) .&. 0xFF
    b3 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
    b4 = fromIntegral $ x .&. 0xFF

xdrReadUInt :: XDRReadS Word32
xdrReadUInt (b1:b2:b3:b4:bs) = (x,bs) where
    x = (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|. 
	    (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)
xdrReadUInt bs = error $ "xdrReadUInt " ++ show bs


putWord32 h x = do
    putRaw h $ xdrWriteUInt x []

readWord32 h = do
    xs <- readRaw h 4 
    return $ fst $ xdrReadUInt xs

galeEncodeString :: String -> [Word8]
galeEncodeString cs = concatMap (f . ord) cs where
    f x = (b1:b2:[]) where 
	b1 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
	b2 = fromIntegral $ x .&. 0xFF

galeDecodeString :: [Word8] -> String
galeDecodeString [] = [] 
galeDecodeString (b1:b2:xs) = (chr $ (fromIntegral b2) .|. (fromIntegral b1 `shiftL` 8)) : galeDecodeString xs

--------------------
-- Category routines
--------------------

catfixes = [ ("/", ".|"), (".", "/"), (":", "..") ]

categoryO2N ('@':r) = con (init c) ++ "@" ++ d where
    (d,'/':'u':'s':'e':'r':'/':c) = span (/= '/') r
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(x,y) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 
    
catParseNew :: String -> Category
catParseNew cs = (c,tail d) where
    (c,d) = span (/= '@') cs
    
catParseOld :: String -> Category
catParseOld ('@':r) = (con (init c),d) where
    (d,'/':'u':'s':'e':'r':'/':c) = span (/= '/') r
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(x,y) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 

catShowOld :: Category -> String
catShowOld (c,d) = "@" ++ d ++ "/user/" ++ con c where 
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(y,x) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 

catShowNew :: Category -> String
catShowNew (x,y) = x ++ "@" ++ y

------------------
-- Examining puffs
------------------

getFragmentString :: Puff -> String -> Maybe String
getFragmentString (Puff _ frags) s = case lookup s frags of
    Just (FragmentText s) -> Just s
    _ -> Nothing

getFragmentStrings :: Puff -> String -> [String]
getFragmentStrings (Puff _ frags) s = [v|(n,FragmentText v) <- frags, n == s]
